package gov.usgs.earthquake.nshmp.netcdf.www;

import java.nio.file.Path;

import gov.usgs.earthquake.nshmp.gmm.NehrpSiteClass;
import gov.usgs.earthquake.nshmp.netcdf.NetcdfDataFilesGroundMotions;
import gov.usgs.earthquake.nshmp.netcdf.www.Metadata.ServiceResponseMetadata;
import gov.usgs.earthquake.nshmp.netcdf.www.Request.RequestDataSiteClass;
import gov.usgs.earthquake.nshmp.www.NshmpMicronautServlet;
import gov.usgs.earthquake.nshmp.www.ResponseBody;

import io.micronaut.context.annotation.Value;
import io.micronaut.context.event.StartupEvent;
import io.micronaut.core.annotation.Nullable;
import io.micronaut.http.HttpRequest;
import io.micronaut.http.HttpResponse;
import io.micronaut.http.MediaType;
import io.micronaut.http.annotation.Controller;
import io.micronaut.http.annotation.Get;
import io.micronaut.http.annotation.PathVariable;
import io.micronaut.http.annotation.QueryValue;
import io.micronaut.runtime.event.annotation.EventListener;
import io.swagger.v3.oas.annotations.Hidden;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.media.Content;
import io.swagger.v3.oas.annotations.media.ExampleObject;
import io.swagger.v3.oas.annotations.media.Schema;
import io.swagger.v3.oas.annotations.responses.ApiResponse;
import io.swagger.v3.oas.annotations.tags.Tag;
import jakarta.inject.Inject;

/**
 * Micronaut controller for getting ground motions from a NetCDF file.
 *
 * @see NetcdfService
 *
 * @author U.S. Geological Survey
 */
@Tag(name = "Web Service")
@Controller("/spectra")
public class NetcdfController {

  @Inject
  private NshmpMicronautServlet servlet;

  @Value("${nshmp-ws-static.netcdf-path}")
  Path netcdfPath;

  @Value("${nshmp-ws-static.aashto-year}")
  int aashtoYear;

  NetcdfServiceGroundMotions service;

  /**
   * Read in data type and return the appropriate service to use.
   */
  @EventListener
  void startup(StartupEvent event) {
    var netcdfDataFiles = new NetcdfDataFilesGroundMotions(netcdfPath, aashtoYear);
    service = new NetcdfServiceGroundMotions(netcdfDataFiles);
  }

  /**
   * GET method to return static curves using slash delimited.
   *
   * @param request The HTTP request
   * @param longitude Longitude of site, in decimal degrees
   * @param latitude Latitude of site, in decimal degrees
   * @param siteClass Site class
   * @param format Optional - Response return type: CSV or JSON. Default: JSON
   */
  @Operation(
      summary = "Returns risk-targeted design response spectra from a slash based call",
      description = "### Returns a risk-targeted design response spectrum for a " +
          "user-specified latitude, longitude, and site class.\n" +

          "Enter the latitude and longitude, select the site class " +
          "and format (optional), and press `Execute`.\n" +

          "### Service Response Types\n" +
          "The web service can respond in JSON or CSV format.\n" +
          "<br><br>" +
          "Choose the format type by adding a `format` query to the service call:" +
          "`?format=JSON` or `?format=CSV`\n" +
          "<br><br>" +
          "Default: `JSON`\n\n" +

          "### Service Call Pattern\n" +
          "This service call is slashed delimited with pattern: " +
          "`/spectra/{longitude}/{latitude}/{siteClass}`\n" +
          "<br><br>" +
          "Example: `/spectra/-118/34/A`\n" +
          "<br><br>" +
          "CSV Example: `/spectra/-118/34/A?format=CSV`\n",
      operationId = "aashto-slash")
  @ApiResponse(
      description = "Spatially interpolates data from https://doi.org/10.5066/P9Z206HY",
      responseCode = "200",
      content = {
          @Content(
              mediaType = MediaType.APPLICATION_JSON,
              schema = @Schema(implementation = Response.class)),
          @Content(
              mediaType = MediaType.TEXT_CSV,
              examples = @ExampleObject(
                  name = "CSV",
                  value = "AASHTO-2023 Web Service\n" +
                      "\n" +
                      "Longitude,-105.0,Latitude,40.0,SiteClass,BC\n" +
                      "Period (s),0.0,0.01,0.02,0.03,0.05,0.075, ...\n" +
                      "Spectral Acceleration (g),0.073,0.08,0.12,0.14,0.18, ...\n"))
      })
  @Get(uri = "/{longitude}/{latitude}/{siteClass}{?format}",
      produces = { MediaType.APPLICATION_JSON, MediaType.TEXT_CSV })
  public HttpResponse<String> doGetSlashBySiteFormat(
      HttpRequest<?> request,
      @Schema(required = true) @PathVariable Double latitude,
      @Schema(required = true) @PathVariable Double longitude,
      @Schema(required = true,
          implementation = NehrpSiteClass.class) @PathVariable NehrpSiteClass siteClass,
      @Schema(required = false,
          defaultValue = "JSON") @QueryValue @Nullable ResponseFormat format) {
    var query = new Query(longitude, latitude, siteClass, format);
    return service.handleServiceCall(request, query);
  }

  /**
   * GET method to return hazard curves using slash delimited.
   *
   * @param request The HTTP request
   * @param longitude The longitude of the site
   * @param latitude Latitude of the site
   * @param format Optional - Response return type: CSV or JSON. Default: JSON
   */
  @Hidden
  @Get(uri = "/{longitude}/{latitude}{?format}",
      produces = { MediaType.APPLICATION_JSON, MediaType.TEXT_CSV })
  public HttpResponse<String> doGetSlashFormat(
      HttpRequest<?> request,
      @Schema(required = true) @PathVariable Double longitude,
      @Schema(required = true) @PathVariable Double latitude,
      @Schema(required = false,
          defaultValue = "JSON") @QueryValue @Nullable ResponseFormat format) {
    return doGetSlashBySiteFormat(request, latitude, longitude, null, format);
  }

  /**
   * GET method to return a static curve using URL query.
   *
   * @param request The HTTP request
   * @param longitude The longitude of the site
   * @param latitude Latitude of the site
   * @param siteClass The site class (optional)
   * @param format Optional - Response return type: CSV or JSON. Default: JSON
   */
  @Hidden
  @Operation(
      summary = "Returns risk-targeted design response spectra from a query based call",
      description = "### Returns a risk-targeted design response spectrum for a " +
          "user-specified latitude, longitude, and site class.\n" +

          "Enter the latitude and longitude select the site class, and press `Execute`.\n" +

          "### Service call pattern\n" +
          "This service call is query based with pattern: " +
          "`/spectra?longitude={number}&latitude={number}&siteClass={string}`\n" +
          "<br><br>" +
          "Example: `/spectra?longitude=-118&latitude=34&siteClass=A`",
      operationId = "aashto-query")
  @ApiResponse(
      description = "Spatially interpolates data from https://doi.org/10.5066/P9Z206HY",
      responseCode = "200",
      content = @Content(
          mediaType = MediaType.APPLICATION_JSON,
          schema = @Schema(implementation = Response.class)))
  @Get(uri = "{?longitude,latitude,siteClass,format}",
      produces = { MediaType.APPLICATION_JSON, MediaType.TEXT_CSV })
  public HttpResponse<String> doGet(
      HttpRequest<?> request,
      @Schema(required = true) @QueryValue @Nullable Double latitude,
      @Schema(required = true) @QueryValue @Nullable Double longitude,
      @QueryValue @Nullable NehrpSiteClass siteClass,
      @Schema(defaultValue = "JSON") @QueryValue @Nullable ResponseFormat format) {
    var query = new Query(longitude, latitude, siteClass, format);
    return service.handleServiceCall(request, query);
  }

  // For Swagger schema
  private static class Response
      extends
      ResponseBody<RequestDataSiteClass, ResponseData<ServiceResponseMetadata>> {}
}
