package gov.usgs.earthquake.nshmp.netcdf.www;

import static gov.usgs.earthquake.nshmp.netcdf.www.NetcdfWsUtils.GSON;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;
import java.util.stream.Collectors;

import gov.usgs.earthquake.nshmp.Text;
import gov.usgs.earthquake.nshmp.Text.Delimiter;
import gov.usgs.earthquake.nshmp.gmm.NehrpSiteClass;
import gov.usgs.earthquake.nshmp.netcdf.Netcdf;
import gov.usgs.earthquake.nshmp.netcdf.www.Metadata.ServiceResponseMetadata;
import gov.usgs.earthquake.nshmp.netcdf.www.Query.Service;
import gov.usgs.earthquake.nshmp.netcdf.www.Request.RequestData;
import gov.usgs.earthquake.nshmp.netcdf.www.Request.RequestDataSiteClass;
import gov.usgs.earthquake.nshmp.www.ResponseBody;

import io.micronaut.http.HttpRequest;
import io.micronaut.http.HttpResponse;

/**
 * Abstract service handler for {@code NetcdfController}.
 *
 * @see NetcdfController
 *
 * @author U.S. Geological Survey
 */
public abstract class NetcdfService<T extends Query> {

  protected static final Logger LOGGER = Logger.getLogger(NetcdfService.class.getName());

  Netcdf<?> netcdf;

  protected NetcdfService(Netcdf<?> netcdf) {
    this.netcdf = netcdf;
  }

  /**
   * Returns the metadata response.
   *
   * @param httpRequest The HTTP request
   */
  abstract ResponseBody<String, Metadata<T>> getMetadataResponse(HttpRequest<?> httpRequest);

  /**
   * Returns the service name
   */
  abstract String getServiceName();

  /**
   * Returns the source model
   */
  abstract SourceModel getSourceModel();

  /**
   * Returns the netcdf object associated with the specific data type.
   */
  abstract Netcdf<?> netcdf();

  /**
   * Returns the static curves at a specific location.
   *
   * @param <T> The response type
   * @param request The request data
   * @param url The URL for the service call
   */
  abstract <U> ResponseBody<RequestData, U> processCurves(RequestData request, String url);

  /**
   * Returns the static curves at a specific location and site class.
   *
   * @param <T> The response type
   * @param request The request data
   * @param url The URL for the service call
   */
  abstract <U> ResponseBody<RequestDataSiteClass, U> processCurvesSiteClass(
      RequestDataSiteClass request,
      String url);

  /**
   * Process the service request and returns the string response.
   *
   * @param httpRequest The HTTP request
   * @param query The HTTP query
   * @param service The NetCDF service
   */
  abstract String processRequest(
      HttpRequest<?> httpRequest,
      T query,
      Service service);

  /**
   * Handler of HTTP request, returns the service response.
   *
   * @param httpRequest The HTTP request
   * @param query The service query
   */
  public HttpResponse<String> handleServiceCall(
      HttpRequest<?> httpRequest,
      T query) {
    try {
      var url = httpRequest.getUri().toString();
      LOGGER.info("Request: " + url);
      LOGGER.fine("Query:\n" + GSON.toJson(query));

      if (query.longitude == null && query.latitude == null) {
        return metadata(httpRequest);
      }
      var service = getService(query);
      var response = processRequest(httpRequest, query, service);
      LOGGER.fine("Result:\n" + response);
      return HttpResponse.ok(response);
    } catch (Exception e) {
      var url = httpRequest.getUri().toString();
      return NetcdfWsUtils.handleError(e, getServiceName(), url);
    }
  }

  public Service getService(T query) {
    if (query.siteClass != null) {
      return Service.CURVES_BY_SITE_CLASS;
    } else {
      return Service.CURVES;
    }
  }

  <U extends ServiceResponseMetadata> String toCsvResponse(
      RequestData requestData,
      ResponseData<U> responseData) {
    return toCsv(requestData, responseData);
  }

  <U extends ServiceResponseMetadata> String toCsvResponseFromList(
      RequestData requestData,
      List<ResponseData<U>> responseData) {
    var response = responseData.stream().map(data -> {
      return toCsv(requestData, data);
    }).collect(Collectors.joining("\n\n"));

    return response;
  }

  <U extends RequestData, V extends ServiceResponseMetadata> String toResponse(
      ResponseBody<U, ResponseData<V>> serviceResponse) {
    if (serviceResponse.getRequest().format == ResponseFormat.CSV) {
      var csvResponse = toCsvResponse(serviceResponse.getRequest(), serviceResponse.getResponse());
      return String.format("%s\n\n%s", getServiceName(), csvResponse);
    } else {
      return NetcdfWsUtils.GSON.toJson(serviceResponse);
    }
  }

  <U extends RequestData, V extends ServiceResponseMetadata> String toResponseFromList(
      ResponseBody<U, List<ResponseData<V>>> serviceResponse) {
    if (serviceResponse.getRequest().format == ResponseFormat.CSV) {
      var csvResponse =
          toCsvResponseFromList(serviceResponse.getRequest(), serviceResponse.getResponse());
      return String.format("%s\n\n%s", getServiceName(), csvResponse);
    } else {
      return NetcdfWsUtils.GSON.toJson(serviceResponse);
    }
  }

  private HttpResponse<String> metadata(HttpRequest<?> httpRequest) {
    var svcResponse = getMetadataResponse(httpRequest);
    var response = GSON.toJson(svcResponse);
    LOGGER.fine("Result:\n" + response);
    return HttpResponse.ok(response);
  }

  private <U extends RequestData, V extends ServiceResponseMetadata> String toCsv(
      U requestData,
      ResponseData<V> responseData) {
    var request = responseData.metadata.toRequestMetadata(requestData);

    List<Object> xs = new ArrayList<>();
    xs.add(responseData.metadata.xLabel);
    xs.addAll(responseData.data.xValues().boxed().collect(Collectors.toList()));

    List<Object> ys = new ArrayList<>();
    ys.add(responseData.metadata.yLabel);
    ys.addAll(responseData.data.yValues().boxed().collect(Collectors.toList()));

    var lines = new ArrayList<String>();
    lines.add(request.toCsv());
    lines.add(Text.join(xs, Delimiter.COMMA));
    lines.add(Text.join(ys, Delimiter.COMMA));

    return lines.stream().collect(Collectors.joining(Text.NEWLINE));
  }

  static class SourceModel {
    public final String name;
    public final Map<NehrpSiteClass, Double> siteClasses;

    SourceModel(Netcdf<?> netcdf) {
      name = netcdf.netcdfData().scienceBaseMetadata().label;
      siteClasses = netcdf.netcdfData().vs30Map();
    }
  }
}
